(* =================================================================== *)
(** === DÉBUT CODE FOURNI PARTIE 1 === *)
(** ========= NE PAS MODIFIER ======== *)

module Tag : sig
  (** Le type est abstrait, ce qui empêche de créer des tags directement.
      Il faut utiliser les fonctions de l'API ci-dessous. *)
  type t

  (** on peut obtenir l'entier unique associé à chaque tag *)
  val to_int : t -> int

  (** tag par défaut, pour la Partie 1 *)
  val default : t

  (** créer un nouveau tag, pour la Partie 3 *)
  val create : unit -> t
end = struct
  type t = int 

  let to_int x = x

  let default = 0 

  let cpt = ref 0

  let create () =
    cpt := !cpt + 1;
    !cpt 
end

(* ------------------------------------------------------------------- *)
(** un terme *)
type term = Tfun of Tag.t * string * term list

(* ------------------------------------------------------------------- *)
(** un terme avec des variables *)
type rule_term =
  | Rvar of int (** on représente chaque variable par un entier *)
  | Rfun of string * rule_term list

(** une règle de réécriture *)
type rule = rule_term * rule_term
                
(* ------------------------------------------------------------------- *)
(** Fonctions de lecture depuis les fichiers.  

    On utilisera seulement les fonctions [read_terms],
    [read_rule_terms] et [read_rules] comme spécifié par l'énoncé.

    Il n'est pas nécessaire de lire le code de ces fonctions. *)

let ident x = x

let parse_rparen chan : unit = Scanf.bscanf chan " ) " ()
  
let rparen chan : bool =
  try parse_rparen chan; true with
  | Scanf.Scan_failure _ -> false

let comma chan : bool =
  try Scanf.bscanf chan " , " (); true with
  | Scanf.Scan_failure _ -> false

(* ------------------------------------------------------------------- *)
let parse_term (str : string) : term =
  let chan = Scanf.Scanning.from_string str in

  let rec parse1 () : term =
    let f = Scanf.bscanf chan " %[a-zA-Z0-9] ( " ident in
    let terms = parse_args () in
    Tfun (Tag.default, f, terms)
                                 
  and parse_args () : term list =
    if rparen chan then [] else parse_list ()

  and parse_list () : term list =
    let t = parse1 () in
    if comma chan then t :: parse_list () else ( parse_rparen chan; [t] )
  in
  parse1 () 

(* ------------------------------------------------------------------- *)
let parse_rule_term chan : rule_term =
   (* call [parse_var] before [parse_fun] because of integers *)
  let rec parse1 () : rule_term =
    try parse_var () with Scanf.Scan_failure _ -> parse_fun ()
                                                    
  and parse_fun () : rule_term =
    let f = Scanf.bscanf chan " %[a-zA-Z0-9] ( " ident in
    let terms = parse_args () in
    Rfun (f, terms)

  and parse_var () : rule_term =
    Rvar (Scanf.bscanf chan " %d " ident)

  and parse_args () : rule_term list =
    if rparen chan then [] else parse_list ()

  and parse_list () : rule_term list =
    let t = parse1 () in
    if comma chan then t :: parse_list () else ( parse_rparen chan; [t] )
  in
  parse1 () 

let parse_rule_term_str (str : string) : rule_term =
  parse_rule_term (Scanf.Scanning.from_string str)

(* ------------------------------------------------------------------- *)
let parse_rule (str : string) : rule =
  let chan = Scanf.Scanning.from_string str in
  Scanf.bscanf chan " %r -> %r "
    parse_rule_term
    parse_rule_term
    (fun left right -> left, right)

(* ------------------------------------------------------------------- *)
(** parse les règles dans le fichier [filename] *)
let read_rules (filename : string) : rule list =
  let file = Stdlib.open_in filename in
  let rec doit () =
    try
      let str = Stdlib.input_line file in
      if String.trim str = "" || String.starts_with ~prefix:"#" str
      then doit ()
      else parse_rule str :: doit () 
    with End_of_file -> []
  in
  doit () 

(* ------------------------------------------------------------------- *)
(** parse les termes de règle dans le fichier [filename] *)
let read_rule_terms (filename : string) : rule_term list =
  let file = Stdlib.open_in filename in
  let rec doit () =
    try
      let str = String.trim (Stdlib.input_line file) in
      if String.trim str = "" || String.starts_with ~prefix:"#" str
      then doit ()
      else parse_rule_term_str str :: doit ()
    with End_of_file -> []
  in
  doit () 

(* ------------------------------------------------------------------- *)
(** parse les termes dans le fichier [filename] *)
let read_terms (filename : string) : term list =
  let file = Stdlib.open_in filename in
  let rec doit () =
    try
      let str = String.trim (Stdlib.input_line file) in
      if String.trim str = "" || String.starts_with ~prefix:"#" str
      then doit ()
      else parse_term str :: doit ()
    with End_of_file -> []
  in
  doit () 

(** ==== FIN CODE FOURNI PARTIE 1 ==== *)
(* =================================================================== *)
(** == DÉBUT DE VOTRE CODE PARTIE 1 == *)






(* ******************************* *)
(* VOTRE CODE ICI POUR LA PARTIE 1 *)
(* ******************************* *)






(* =================================================================== *)
(** === DÉBUT CODE FOURNI PARTIE 3 === *)
(** ========= NE PAS MODIFIER ======== *)

(* ------------------------------------------------------------------- *)
(** Le module [HashCons]. Seuls le type [HashCons.hashtbl] et la
    fonction [HashCons.fun_hashcons] sont utilisables à l'extérieur du
    module.

    Les autres fonctions (par exemple [hash_for_hashconsing]) sont
    cachées par la signature du module. *)
module HashCons : sig
  
  (** Type pour la table de hachage implémentant la structure H. *)
  type hashtbl = { table : term list array; size : int }

  val fun_hashcons : string -> term list -> term
    
end = struct
  let get_tag (t : term) = match t with Tfun (tag,_,_) -> tag

  let hash_for_hashconsing (t : term) : int =
    let s, args = match t with Tfun (_, s, args) -> s, args in
    let h = ref (Hashtbl.hash s) in
    let i = ref 1 in
    List.iter (fun arg ->
        h := !h + ( !i * Tag.to_int (get_tag arg));
        i := !i + 1;
      ) args;
    !h

  let equal_for_hashconsing t1 t2 =
    match t1, t2 with
    | Tfun (_, f1, l1), Tfun (_, f2, l2) ->
      f1 = f2 &&
      List.length l1 = List.length l2 &&
      List.for_all2 (fun x1 x2 -> get_tag x1 = get_tag x2) l1 l2
      (** Documentation : la fonction [List.for_all2] est de signature
          [val for_all2 : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool] 
          et se comporte comme [for_all], mais pour un prédicat à deux 
          arguments. Plus précisement,
          [for_all2 f [a1; ...; an] [b1; ...; bn]] retourne
          [(f a1 b1) && (f a2 b2) && ... && (f an bn)] si les listes sont
          de même longueur, et l'exception [Invalid_argument] sinon. *)
  
  (* ------------------------------------------------------------------- *)
  (** Type pour la table de hachage implémentant la structure H *)
  type hashtbl = { table : term list array; size : int }

  let create n = { table = Array.make n []; size = n; }

  (* ------------------------------------------------------------------- *)
  let add (htbl : hashtbl) (t : term) : unit =
    let hash = (hash_for_hashconsing t) mod htbl.size in
    let l = htbl.table.(hash) in
    htbl.table.(hash) <- (t :: l) 
      
  let find (htbl : hashtbl) (t : term) : term =
    let hash = (hash_for_hashconsing t) mod htbl.size in
    let l = htbl.table.(hash) in
    List.find (equal_for_hashconsing t) l

  (* ------------------------------------------------------------------- *)
  let hashtbl = create 1000000

  (** construit le terme [f(terms)] hashconsé, à condition que les
      termes dans [args] soient déjà hash-consés *)
  let fun_hashcons (f : string) (args : term list) : term =
    let t = Tfun (Tag.default, f, args) in
    try find hashtbl t 
    with Not_found -> 
      let t = Tfun (Tag.create (), f, args) in
      add hashtbl t;
      t
end

(** ==== FIN CODE FOURNI PARTIE 3 ==== *)
(* =================================================================== *)
(** == DÉBUT DE VOTRE CODE PARTIE 3 == *)





(* ******************************* *)
(* VOTRE CODE ICI POUR LA PARTIE 3 *)
(* ******************************* *)
