/* =================================================================== */
/** ======== DÉBUT CODE FOURNI ======= */
/** ========= NE PAS MODIFIER ======== */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <assert.h>
#include <string.h>
#include <stdint.h>

/* ------------------------------------------------------------------- */
/** génère un tableau de nombres pseudo-aléatoires de longueur `length` */
uint32_t* rands(uint32_t u0, uint32_t len){
  uint64_t a = 1400306129;
  uint64_t n = 2038075531;

  uint32_t* res = malloc(len * sizeof(uint32_t));

  res[0] = u0;
  
  for (int i = 1; i < len; i++){
    res[i] = ( a * (uint64_t)res[i-1]) % n;
  }

  return res;
}

/* ------------------------------------------------------------------- */
/** fonction de hachage à utiliser dans une table de hachage de taille `N` */
uint32_t hash(uint32_t x, uint32_t N){
  x = ((x >> 16) ^ x) * 0x45d9f3b;
  x = ((x >> 16) ^ x) * 0x45d9f3b;
  x = (x >> 16) ^ x;
  return ((uint32_t) x) % N;
}

/** ========= FIN CODE FOURNI ======== */
/* =================================================================== */
/** == DÉBUT DE VOTRE CODE PARTIE 3 == */


/* =================================================================== */
/** une entré dans une table de hachage */
struct record_t {
  uint32_t key;
  uint32_t value;
  uint32_t hash;
  bool empty;              /* true: empty; false: used */
};
typedef struct record_t record;

/* ------------------------------------------------------------------- */
/** une table de hachage */
struct htable_t {
  uint32_t max_size;       /* taille maximum de la table de hachage */
  uint32_t size;           /* taille actuelle de la table de hachage*/
  bool resizable;          /* est-ce que la table peut être rehaché */
  record *data;            /* pointeur vers un tableau de [max_size] case */
};
typedef struct htable_t htable;

/* ------------------------------------------------------------------- */
void print_record(htable *tbl, record *r){
  if (r->empty){
    printf("⊥");
  } else {
    printf("{k: %4d, v: %4d, h: %10u}", r->key, r->value, r->hash);
  }
}

/* ------------------------------------------------------------------- */
void print_table(htable *tbl){
  printf("[max_size = %d\n size     = %d\n data =\n", tbl->max_size, tbl->size);

  for(uint32_t i = 0; i < tbl->max_size; i++){
    printf("%4d: ", i);
    print_record(tbl, &tbl->data[i]);
    printf("\n");
  }
  printf("]\n\n");
}


/* ------------------------------------------------------------------- */
void sanity_check(htable *tbl){}

/* void sanity_check(htable *tbl){ */
/*   uint32_t size=0; */
  
/*   for(uint32_t i = 0; i<tbl->max_size; i++){ */
/*     if(!tbl->data[i].empty){ */
/*       size++; */
/*       for(uint32_t j = tbl->data[i].hash; */
/*           j != i; */
/*           j = (j + 1) % tbl->max_size){ */
/*         assert(!tbl->data[j].empty); */
/*       } */
/*     } */
/*   } */
/*   assert(size == tbl->size); */
/* } */


/* ------------------------------------------------------------------- */
record* create_records(uint32_t max_size){
  record *data = malloc(max_size * sizeof(record));

  /* toutes les cellules sont initiallement vide */
  for(uint32_t i=0; i < max_size; i++){
    data[i].empty = true;
  }

  return data;
}

/* ------------------------------------------------------------------- */
htable* create(uint32_t max_size, bool resizable){
  htable *h = malloc(sizeof(htable));
  record *data = create_records(max_size);

  h->max_size=max_size;
  h->size=0;
  h->data = data;
  h->resizable = resizable;
  return h;
}

/* ------------------------------------------------------------------- */
void add0(htable *tbl, uint32_t key, uint32_t val, uint32_t h){
  uint32_t i = 0;
  uint32_t cell;
  record *r;
  
  /* on cherche la première cellule vide ou cellule contenant `key` à
     partir de `h` */
  while(i < tbl->max_size){
    cell = (h + i) % tbl->max_size;
    r    = &(tbl->data[cell]);
    if (r->empty     ) { tbl->size++; break; }
    if (r->key == key) {              break; }
    i++;
  }

  if(i == tbl->max_size) { abort (); }  /* on crash si la tbl est pleine */

  r->key   = key;
  r->value = val;
  r->hash  = h;
  r->empty = false;

  sanity_check(tbl); 
}


/* ------------------------------------------------------------------- */
void rehash(htable *tbl, uint32_t new_max_size){
  record *new_data = create_records(new_max_size);
  record *old_data = tbl->data;
  uint32_t old_max_size = tbl->max_size;

  tbl->max_size=new_max_size;
  tbl->data = new_data;
  tbl->size=0;

  /* toutes les cellules sont initiallement vide */
  for(int i=0; i < old_max_size; i++){
    if (!old_data[i].empty) {
      uint32_t new_hash = hash(old_data[i].key, new_max_size);
      add0(tbl, old_data[i].key, old_data[i].value, new_hash);
    }
  }

  free(old_data);
  sanity_check(tbl);
}

/* ------------------------------------------------------------------- */
void add(htable *tbl, uint32_t key, uint32_t val){
  uint32_t h = hash(key, tbl->max_size);

  add0(tbl,key,val,h);

  if (tbl->resizable && tbl->size > tbl->max_size/4) {
    rehash(tbl, tbl->max_size * 2);
  }
}

/* ------------------------------------------------------------------- */
/** Stock dans `res` l'indice de la cellule contenant `key` dans `tbl`.
    Retourn `true` si l'indice a été trouvé, `false` sinon  */
bool lookup_index(htable *tbl, uint32_t key, uint32_t *res){
  uint32_t h = hash(key,tbl->max_size);
  int i = 0;

  /* on cherche la cellule contenant `key` */
  while(i < tbl->max_size){
    uint32_t index = (h + i) % tbl->max_size;
    if (tbl->data[index].key == key) { *res = index; return true; }
    if (tbl->data[index].empty) { return false; }
    i++;
  }
  return false;
}

/* ------------------------------------------------------------------- */
/** retourne la valeur par défaut `0` si une clé n'est pas dans la table */
uint32_t lookup(htable *tbl, uint32_t key){
  /* on cherche la cellule contenatn `key` */
  uint32_t index;
  if (!lookup_index(tbl, key, &index)) { return 0; }

  return tbl->data[index].value;
}


/* ------------------------------------------------------------------- */
uint32_t count_collisions(htable *tbl){
  uint32_t collisions = 0;
  int i;

  for(i = 0; i < tbl->max_size; i++){
    if(!tbl->data[i].empty &&
       (tbl->data[i].hash != i)) { collisions++; }
  }
  return collisions;
}


/* ------------------------------------------------------------------- */
uint32_t q6(uint32_t u0, uint32_t len){
  uint32_t* res = rands(u0, len);

  htable *d = create(4 * len,false);
  
  for (int i = 0; i < len; i++){
    uint32_t key = res[i] % (10 * len);
    uint32_t val = (17 * (res[i] + i)) % 20;
    add(d,key,val);
  }

  return count_collisions(d);
}
/* ------------------------------------------------------------------- */
uint32_t q7(uint32_t u0, uint32_t len){
  uint32_t* res = rands(u0, len);

  htable *d = create(16,true);
  
  for (int i = 0; i < len; i++){
    uint32_t key = res[i] % (10 * len);
    uint32_t val = (17 * (res[i] + i)) % 20;
    add(d,key,val);
  }

  return count_collisions(d);
}
  
/* ------------------------------------------------------------------- */
int main (int argc, char** argv){
  uint32_t u0 = 1;

  if (argc > 1) {
    int r = sscanf(argv[1], "%d", &u0);
    if (r != 1) {
      printf("Erreur : u0 invalide donné en argument\n");
      exit(1);
    }
  }

  /* ------------------------------------------------------------------- */
  printf("6.a %d\n", q6(u0,      100));
  printf("6.b %d\n", q6(u0,    10000));
  printf("6.c %d\n", q6(u0,  1000000));
  printf("6.d %d\n", q6(u0, 10000000));

  printf("\n");

  /* ------------------------------------------------------------------- */
  printf("7.a %d\n", q7(u0,      100));
  printf("7.b %d\n", q7(u0,    10000));
  printf("7.c %d\n", q7(u0,  1000000));
  printf("7.d %d\n", q7(u0, 10000000));

  
  return 1;
}
