(* =================================================================== *)
(** == DÉBUT CODE FOURNI PARTIE 1 == *)
(** ======== NE PAS MODIFIER ======= *)

module Tag : sig
  (** Le type est abstrait, ce qui empêche de créer des tags directement.
      Il faut utiliser les fonctions de l'API ci-dessous. *)
  type t

  (** on peut obtenir l'entier unique associé à chaque tag *)
  val to_int : t -> int

  (** tag par défaut, pour la Partie 1 *)
  val default : t

  (** créer un nouveau tag, pour la Partie 3 *)
  val create : unit -> t
end = struct
  type t = { tag : int }

  let to_int x = x.tag

  let default = { tag = 0 }

  let cpt = ref 0
  let create () = cpt := !cpt + 1; { tag = !cpt }
end

(* ------------------------------------------------------------------- *)
(** un terme *)
type term = Tfun of Tag.t * string * term list

let tfun s l = Tfun (Tag.default, s, l)

(* ------------------------------------------------------------------- *)
(** un terme avec des variables *)
type rule_term =
  | Rvar of int (** on représente chaque variable par un entier *)
  | Rfun of string * rule_term list

(** une règle de ré-écriture *)
type rule = rule_term * rule_term
                
(* ------------------------------------------------------------------- *)
(** Fonctions de lectures depuis les fichiers.  

    On utilisera seulement les fonctions [read_terms],
    [read_rule_terms] et [read_rules] tel que spécifié par l'énoncé.

    Il n'est pas nécessaire de lire le code de ces fonctions. *)

let ident x = x

let parse_rparen chan : unit = Scanf.bscanf chan " ) " ()
  
let rparen chan : bool =
  try parse_rparen chan; true with
  | Scanf.Scan_failure _ -> false

let comma chan : bool =
  try Scanf.bscanf chan " , " (); true with
  | Scanf.Scan_failure _ -> false

(* ------------------------------------------------------------------- *)
let parse_term (str : string) : term =
  let chan = Scanf.Scanning.from_string str in

  let rec parse1 () : term =
    let f = Scanf.bscanf chan " %[a-zA-Z0-9] ( " ident in
    let terms = parse_args () in
    tfun f terms
                                 
  and parse_args () : term list =
    if rparen chan then [] else parse_list ()

  and parse_list () : term list =
    let t = parse1 () in
    if comma chan then t :: parse_list () else ( parse_rparen chan; [t] )
  in
  parse1 () 

(* ------------------------------------------------------------------- *)
let parse_rule_term chan : rule_term =
   (* call [parse_var] before [parse_fun] because of integers *)
  let rec parse1 () : rule_term =
    try parse_var () with Scanf.Scan_failure _ -> parse_fun ()
                                                    
  and parse_fun () : rule_term =
    let f = Scanf.bscanf chan " %[a-zA-Z0-9] ( " ident in
    let terms = parse_args () in
    Rfun (f, terms)

  and parse_var () : rule_term =
    Rvar (Scanf.bscanf chan " %d " ident)

  and parse_args () : rule_term list =
    if rparen chan then [] else parse_list ()

  and parse_list () : rule_term list =
    let t = parse1 () in
    if comma chan then t :: parse_list () else ( parse_rparen chan; [t] )
  in
  parse1 () 

let parse_rule_term_str (str : string) : rule_term =
  parse_rule_term (Scanf.Scanning.from_string str)

(* ------------------------------------------------------------------- *)
let parse_rule (str : string) : rule =
  let chan = Scanf.Scanning.from_string str in
  Scanf.bscanf chan " %r -> %r "
    parse_rule_term
    parse_rule_term
    (fun left right -> left, right)

(* ------------------------------------------------------------------- *)
(** parse les règles dans le fichier [filename] *)
let read_rules (filename : string) : rule list =
  let file = Stdlib.open_in filename in
  let rec doit () =
    try
      let str = Stdlib.input_line file in
      if String.trim str = "" || String.starts_with ~prefix:"#" str
      then doit ()
      else parse_rule str :: doit () 
    with End_of_file -> []
  in
  doit () 

(* ------------------------------------------------------------------- *)
(** parse les termes de règle dans le fichier [filename] *)
let read_rule_terms (filename : string) : rule_term list =
  let file = Stdlib.open_in filename in
  let rec doit () =
    try
      let str = String.trim (Stdlib.input_line file) in
      if String.trim str = "" || String.starts_with ~prefix:"#" str
      then doit ()
      else parse_rule_term_str str :: doit ()
    with End_of_file -> []
  in
  doit () 

(* ------------------------------------------------------------------- *)
(** parse les termes dans le fichier [filename] *)
let read_terms (filename : string) : term list =
  let file = Stdlib.open_in filename in
  let rec doit () =
    try
      let str = String.trim (Stdlib.input_line file) in
      if String.trim str = "" || String.starts_with ~prefix:"#" str
      then doit ()
      else parse_term str :: doit ()
    with End_of_file -> []
  in
  doit () 

(** === FIN CODE FOURNI PARTIE 1 === *)

let u0 = try int_of_string Sys.argv.(1) with Invalid_argument _ -> 1
  
let () = Format.printf "u0 %d@.@." u0
    
(* =================================================================== *)
(** == Utilitaires == *)

(* ------------------------------------------------------------------- *)
(** affiche un terme *)
let rec pp_term (fmt : Format.formatter) (t : term) =
  let Tfun (_, f, terms) = t in
  let pp_terms =
    Format.pp_print_list
      ~pp_sep:(fun fmt () -> Format.fprintf fmt ",")
      pp_term
  in
  Format.fprintf fmt "%s(%a)" f pp_terms terms

(* ------------------------------------------------------------------- *)
(** affiche un terme de règle *)
let rec pp_rule_term (fmt : Format.formatter) (t : rule_term) =
  match t with
  | Rfun (f, terms) ->
    let pp_terms =
      Format.pp_print_list
        ~pp_sep:(fun fmt () -> Format.fprintf fmt ",")
        pp_rule_term
    in
    Format.fprintf fmt "%s(%a)" f pp_terms terms

  | Rvar i -> Format.fprintf fmt "%d" i

(* ------------------------------------------------------------------- *)
(** affiche une règle *)
let pp_rule (fmt : Format.formatter) ((left, right) : rule) =
  Format.fprintf fmt "@[@[%a@]@ → @[%a@]@]" pp_rule_term left pp_rule_term right

(* ------------------------------------------------------------------- *)
(** tests *)

let tpath = "template/"
let path  = "data/" ^ string_of_int u0 ^ "/"

(* ------------------------------------------------------------------- *)
let rec is_int = function
  | Tfun (_, "s", [t]) -> is_int t
  | Tfun (_, "z", []) -> true
  | _ -> false
    
let rec count = function
  | Tfun (_, "s", [term]) -> 1 + count term
  | Tfun (_, "z", []) -> 0
  | _ -> assert false

(** fonction d'affichage plus concise *)
let rec pp_term (fmt : Format.formatter) (t : term) =
  match t with
  (* ad hoc printer, to help read the rules *)
  | _ when is_int t -> Format.fprintf fmt "%d" (count t)

  | Tfun (_, "true" , []) -> Format.fprintf fmt "⊤"
  | Tfun (_, "false", []) -> Format.fprintf fmt "⊥"
  | Tfun (_, "END"  , []) -> Format.fprintf fmt "⋄"
                               
  | Tfun (_, "p", [term1; term2]) ->
    Format.fprintf fmt "%a :: %a" pp_term term1 pp_term term2

  | Tfun (_, f, terms) ->
    let pp_terms =
      Format.pp_print_list
        ~pp_sep:(fun fmt () -> Format.fprintf fmt ",")
        pp_term
    in
    Format.fprintf fmt "%s(%a)" f pp_terms terms

(* ------------------------------------------------------------------- *)
let rec pp_term_dbg (fmt : Format.formatter) (t : term) =
  match t with
  | Tfun (tag, f, terms) ->
    let pp_terms =
      Format.pp_print_list
        ~pp_sep:(fun fmt () -> Format.fprintf fmt ",")
        pp_term_dbg
    in
    Format.fprintf fmt "%s%s(%a)"
      f (if tag <> Tag.default then ":" ^ string_of_int @@ Tag.to_int tag else "")
      pp_terms terms

(* ------------------------------------------------------------------- *)
let pp_option pp fmt = function
  | None   -> Format.fprintf fmt "None"
  | Some x -> Format.fprintf fmt "Some %a" pp x

(* ------------------------------------------------------------------- *)
let max_var = 10

(* =================================================================== *)
(** Partie 1 *)

module Make (T : sig
    val kind : [`Standard | `Hashconsing]
    val tfun  : string -> term list -> term
    val equal : term -> term -> bool
  end) =
struct

open T

(* ------------------------------------------------------------------- *)
(** Question 1 *)

let n = 104_729
  
let rec hash (t : term) : int =
  let Tfun (_, s, args) = t in
  let h = ref (Hashtbl.hash s mod n) in
  List.iteri (fun i arg ->
      h := (!h + (( (i + 1) * hash arg) mod n)) mod n
    ) args;
  !h

let _res1 =
  if kind = `Standard then begin
    let terms = read_terms (path ^ "q1.txt") in
    Format.printf "1.a %d@." (hash (List.nth terms 0));
    Format.printf "1.b %d@." (hash (List.nth terms 1));
    Format.printf "1.c %d@." (hash (List.nth terms 2));
    Format.printf "1.d %d@." (hash (List.nth terms 3));
    Format.printf "@."
  end

(* ------------------------------------------------------------------- *)
(** Question 2 *)

type subst = term option Array.t

let pp_subst fmt (s : subst) =
  Format.fprintf fmt "@[<v 2>";
  Array.iteri (fun i t ->
      match t with
      | None   -> ()
      | Some t -> Format.fprintf fmt "@[%d -> %a@]@;" i pp_term t
    ) s;
  Format.fprintf fmt "@]"
  
let get (s : subst) (v : int) : term =
  match s.(v) with Some x -> x | _ -> tfun "missing" []

let rec subst (s : subst) ( t : rule_term) : term =
  match t with
  | Rvar v -> get s v
  | Rfun (f, l) -> tfun f (List.map (subst s) l)
    
let _res2 =
  if kind = `Standard then begin
    let rule_terms = read_rule_terms (path ^ "q2.txt") in

    let subst_q2 = Array.make max_var None in
    let (!) = fun x -> Some (parse_term x) in
    subst_q2.(0) <- !"s(z())";
    subst_q2.(1) <- !"z()";
    subst_q2.(2) <- !"h(b(),s(z()))";

    let doit t = hash (subst subst_q2 t) in
    Format.printf "2.a %d@." (doit (List.nth rule_terms 0));
    Format.printf "2.b %d@." (doit (List.nth rule_terms 1));
    Format.printf "2.c %d@." (doit (List.nth rule_terms 2));
    Format.printf "2.d %d@." (doit (List.nth rule_terms 3));
    Format.printf "@."
  end

(* QO1: [subst s t]

   Complexité en temps (tableaux): O(|t|) 
   - accès au tableau de la substitution en O(1) 
   - pas de copie des termes lors de la substitution d'une
     variables.

   Complexité en temps (list d'association): O(|t| * V) 
   - V: nombres de variables dans t (donc V <= |t|)
   - accès au tableau de la substitution en O(V) 
*)

(* ------------------------------------------------------------------- *)
(** Question 3 *)

let find_match (t1 : term) (t2 : rule_term) : subst option =
  let exception NoMatch in
  let no_match () = raise NoMatch in

  let s = Array.make max_var None in

  let rec doit t1 t2 =
    let Tfun (_,f1,l1) = t1 in
    match t2 with
    | Rfun (f2, l2) ->
      if f1 <> f2 then no_match ();
      List.iter2 doit l1 l2

    | Rvar v -> 
      match s.(v) with
      | None   -> s.(v) <- Some t1 
      | Some t ->
        (* on fait attention à bien utiliser le test d'égalité rapide *)
        if not (equal t t1) then no_match ()
  in 
  try doit t1 t2; Some s with NoMatch -> None 

let _res3 =
  if kind = `Standard then begin
    let terms = read_terms (path ^ "q3.txt") in
    let (!) = fun x -> parse_rule_term_str x in
    let doit t r =
      Option.map
        (fun s -> hash (subst s (Rvar 0)) + hash (subst s (Rvar 1)))
        (find_match t r)
    in
    let pp = pp_option (fun fmt d -> Format.fprintf fmt "%d" d) in
    
    Format.printf "3.a %a@." pp (doit (List.nth terms 0) !"eq(0,0)");
    Format.printf "3.b %a@." pp (doit (List.nth terms 1) !"eq(0,0)");
    Format.printf "3.c %a@." pp (doit (List.nth terms 2) !"F(f(0),1,g(1))");
    Format.printf "3.d %a@." pp (doit (List.nth terms 3) !"F(f(0),1,g(1))");
    Format.printf "@."
  end

(* QO2: [find_match t1 t2]

   On regarde d'abord le cas ou [t2] est linéaire, i.e. ou chaque
   variable apparaît au plus une fois dans [t2].

   Complexité en temps (t2 linéaire): O(min(|t1|, |t2|) )
   - [doit] est appelé au plus une fois par symbole 
     dans [t1] (idem pour [t2])
   - toutes les autres opérations sont en O(1) car le teste 
     d'égalité dans le cas ou une variable est vu deux fois 
     n'a jamais lieu.

   Complexité en temps (cas général): O(|t1| * min(|t1|, |t2|))
   - chaque test d'égalité est en temps O(|t1|) (car on compare 
     l'égalité de deux sous-termes de [t2])
   - il y a au plus un test d'égalité par feuille de [t1].
     Idem pour [t2]. Donc au plus min(|t1|, |t2|) tests.
*)

(* ------------------------------------------------------------------- *)
(** Question 4 *)
 
let rewrite1_basic (t : term) (rules : rule list) : term option =
  let found = ref false in

  (* cherche à appliquer une règle dans [t] *)
  let rec doit t =
    if !found then t else
      (* on cherche une règle s'appliquant en tête *)
      match t with
      | Tfun (_, f, l) as t ->
        let t = (* itére sur les règles *)
          List.find_map (fun (rl, rr) ->
              match find_match t rl with
              | None   -> None

              (* [s] tel que [t = subst s rl], on renvoit [subst s rr] *)
              | Some (s : subst) -> Some (subst s rr)
            ) rules
        in
        match t with
        | Some t -> found := true; t   (* match trouvé en position de tête *)

        (* pas de match en position de tête dans [t], on récurse *)
        | None -> tfun f (List.map doit l)
  in
  let t = doit t in
  if !found then Some t else None

let _res4 =
    if kind = `Standard then begin
      let terms = read_terms (path ^ "tq4.txt") in
      let rules = read_rules (path ^ "rq4.txt") in
      let doit t =
        Option.map (fun t -> hash t) (rewrite1_basic t rules)
      in
      let pp = pp_option (fun fmt d -> Format.fprintf fmt "%d" d) in

      Format.printf "4.a %a@." pp (doit (List.nth terms 0));
      Format.printf "4.b %a@." pp (doit (List.nth terms 1));
      Format.printf "4.c %a@." pp (doit (List.nth terms 2));
      Format.printf "4.d %a@." pp (doit (List.nth terms 3));
      Format.printf "@."
end

let get_tag (t : term) = match t with Tfun (tag,_,_) -> tag

(* ------------------------------------------------------------------- *)
let rewrite1_hashconsing (t : term) (rules : rule list) : term option =
  let seen = Hashtbl.create 100000 in
  let found = ref false in
  
  (* cherche à appliquer une règle dans [t] *)
  let rec doit t =
    if !found || Hashtbl.mem seen (get_tag t) then t else begin
      Hashtbl.add seen (get_tag t) ();
      (* on cherche une règle s'appliquant en tête *)
      match t with
      | Tfun (_, f, l) as t ->
        let t = (* itére sur les règles *)
          List.find_map (fun (rl, rr) ->
              match find_match t rl with
              | None   -> None

              (* [s] tel que [t = subst s rl], on renvoit [subst s rr] *)
              | Some (s : subst) -> Some (subst s rr)
            ) rules
        in
        match t with
        | Some t -> found := true; t   (* match trouvé en position de tête *)

        (* pas de match en position de tête dans [t], on récurse *)
        | None -> tfun f (List.map doit l)
    end
  in
  let t = doit t in
  if !found then Some t else None
   
let rewrite1 (t : term) (rules : rule list) : term option =
  if get_tag t = Tag.default
  then rewrite1_basic       t rules
  else rewrite1_hashconsing t rules
  
(* QO3 et QO4: [rewrite1_basic t rules]

   - N règles dans [rules] avec au plus V variables différentes.
   - chaque règle [(l,r)] de [rules] est L-linéaire à droite, et |r| < R

   Complexité en temps: O(|t|^3 * N)
   - pour chaque sous-terme de [t] et règle dans [rules], donc |t| * N fois
   - on apple [find_match], en temps O(|t|^2) (on simplifie la borne précédente).

   Complexité en espace: O(|t| + |r| + V * (L - 1) * |t|)
   - on raisonne en regardant ce qu'on ajoute au terme [t] initial
   - supposons que l'on ré-écrive la règle [(l,r)]
   - on enlève |l| noeuds et on en ajoute |r|, donc ajoute 
     au plus |r| noeuds
   - de plus, pour chaque variable [v], on a au plus [L-1] copie 
     d'un sous-terme de [t], donc un ajoute au plus [V * (L-1) * |t|] noeuds

   En supposant l'ensemble des règles fixé:
   - si L = 1, alors on a |t| + |r| = O(|t|) 
   - si L > 1, alors on a O (C * |t|) pour un certain entier C > 0
*)

(* ------------------------------------------------------------------- *)
(** Question 5 *)

let rewriteN (t : term) ~(rules : rule list) : term * int =
  let rec doit (t : term) (step : int) =
    (* let () = Format.eprintf "  @[%a@] → @." pp_term t in *)
    match rewrite1 t rules with
    | None   -> (t, step)
    | Some t -> doit t (step + 1)
  in
  doit t 0

let _res5 =
  if kind = `Standard then begin
    let terms = read_terms (path ^ "tq5.txt") in
    let rules = read_rules (path ^ "rq5.txt") in
    let doit t =
      let t', steps = rewriteN t ~rules in
      assert (steps < n);
      steps
    in
    let pp fmt d = Format.fprintf fmt "%d" d in

    Format.printf "5.a %a@." pp (doit (List.nth terms 0));
    Format.printf "5.b %a@." pp (doit (List.nth terms 1));
    Format.printf "5.c %a@." pp (doit (List.nth terms 2));
    Format.printf "5.d %a@." pp (doit (List.nth terms 3));
    Format.printf "@."
  end
  
(* QO5: [rewriteN t ~rules ~n]

   - si les règles sont L-linéaire à droite, en supposant l'ensemble
     de règles fixé

   Complexité en espace (L = 1): O(|t| + n)
   - chaque ré-écriture ajoute au plus un nombre constant de noeuds

   Complexité en temps (L = 1): O(n * (|t| + n)^3)
   - en utilisant le fait que le i-éme terme dans la séquence 
     de ré-écriture est de taille borné par |t| + i * k 
     (pour une certaine constante k), on borne la complexité par:

          sum_{i <= n} (|t| + i * k)^3
       <= n * (|t| + n)^3

   Complexité en espace (L > 1): O(C^n * |t|)
   - chaque ré-écriture multiplie la taille du term par au-plus C

   Complexité en temps (L > 1): O(n * |t|^3 * C^{3 * n})
   - exponentielle
*)

end (** end of module [Make] *)
    

(* =================================================================== *)
(** Partie 2 
    Noter que certaines des fonctions sont implémentées dans le 
    module [Make] ci-dessus. *)

(* =================================================================== *)
(** == DÉBUT CODE FOURNI PARTIE 2 == *)
(** ======== NE PAS MODIFIER ======= *)

(* ------------------------------------------------------------------- *)
module HashCons : sig
  (** Type pour la table de hachage implémentant la structure H *)
  type hashtbl = { table : term list Array.t; size  : int }

  val fun_hashcons : string -> term list -> term 
end = struct
  let get_tag (t : term) = match t with Tfun (tag,_,_) -> tag

  let hash_for_hashconsing (t : term) : int =
    let s, args = match t with Tfun (_, s, args) -> s, args in
    let h = ref (Hashtbl.hash s) in
    let i = ref 1 in
    List.iter (fun arg ->
        h := !h + ( !i * Tag.to_int (get_tag arg));
        i := !i + 1;
      ) args;
    !h

  let equal_for_hashconsing t1 t2 =
    match t1, t2 with
    | Tfun (_, f1, l1), Tfun (_, f2, l2) ->
      f1 = f2 &&
      List.length l1 = List.length l2 &&
      List.for_all2 (fun x1 x2 -> get_tag x1 = get_tag x2) l1 l2
      (** Documentation : la fonction [List.for_all2] est de signature
          [val for_all2 : ('a -> 'b -> bool) -> 'a list -> 'b list -> bool] 
          et se comporte comme [for_all], mais pour un prédicat à deux 
          arguments. Plus précisement,
          [for_all2 f [a1; ...; an] [b1; ...; bn]] retourne
          [(f a1 b1) && (f a2 b2) && ... && (f an bn)] si les listes sont
          de même longueurs, et l'exception [Invalid_argument] sinon. *)
  
  (* ------------------------------------------------------------------- *)
  (** Type pour la table de hachage implémentant la structure H *)
  type hashtbl = { table : term list Array.t; size  : int }

  let create n = { table = Array.make n []; size = n; }

  (* ------------------------------------------------------------------- *)
  let add (htbl : hashtbl) (t : term) : unit =
    let hash = (hash_for_hashconsing t) mod htbl.size in
    let l = htbl.table.(hash) in
    htbl.table.(hash) <- (t :: l) 
      
  let find (htbl : hashtbl) (t : term) : term =
    let hash = (hash_for_hashconsing t) mod htbl.size in
    let l = htbl.table.(hash) in
    List.find (equal_for_hashconsing t) l

  (* ------------------------------------------------------------------- *)
  let hashtbl = create 1000000

  (** construit le terme [f(terms)] hashconsé, à condition que les
      termes dans [args] soient déjà hash-consées *)
  let fun_hashcons (f : string) (args : term list) : term =
    let t = Tfun (Tag.default, f, args) in
    try find hashtbl t 
    with Not_found -> 
      let t = Tfun (Tag.create (), f, args) in
      add hashtbl t;
      t
end

(** == FIN CODE FOURNI == *)

(* ------------------------------------------------------------------- *)
(** Question 8 *)
  
(** hashconse un terme arbitraire *)
let rec hashcons (t : term) : term =
  match t with
  | Tfun (_, f, terms) -> HashCons.fun_hashcons f (List.map hashcons terms)

(* ------------------------------------------------------------------- *)
(** compte le nombre de noeud dans un terme *)
let size (t : term) : int =
  let cpt = ref 0 in

  let rec doit t : unit =
    match t with
    | Tfun (tag, _, terms) ->
      incr cpt;
      List.iter doit terms
  in
  doit t;
  !cpt

(* ------------------------------------------------------------------- *)
(** compte le nombre de noeud dans un terme hashconsé *)
let size_mem (t : term) : int =
  (* ensemble des terms déjà vu, stocké par leur tag *)
  let seen = Hashtbl.create 100000 in
  let cpt = ref 0 in

  let rec doit t : unit =
    match t with
    | Tfun (tag, _, terms) ->
      assert (tag <> Tag.default);
      let tag = Tag.to_int tag in
      (* on ne recompte pas un noeud déjà compté *)
      if not (Hashtbl.mem seen tag) then
        begin
          Hashtbl.add seen tag ();
          incr cpt;
          List.iter doit terms
        end;
  in
  doit t;
  !cpt

(* ------------------------------------------------------------------- *)
(** numerical results post-poned below *)

let get_tag (t : term) = match t with Tfun (tag,_,_) -> tag

(* ------------------------------------------------------------------- *)
let size_smart (t : term) : int =
  if get_tag t = Tag.default then size t else size_mem t
                    
(* ------------------------------------------------------------------- *)
(** compte la taille qu'aurait un terme hashconsé totalement déplié *)
let size_virtual (t : term) : int =
  (* taille virtuelle des terms déjà vu, stocké par leur tag *)
  let htbl = Hashtbl.create 100000 in

  let rec size t : int =
    match t with
    | Tfun (tag, _, terms) ->
      assert (tag <> Tag.default);
      let tag = Tag.to_int tag in
      if Hashtbl.mem htbl tag then
        Hashtbl.find htbl tag
      else
        let s = List.fold_left (fun s t' -> s + size t') 1 terms in
        let () = Hashtbl.add htbl tag s in
        s
  in
  size t

(* ------------------------------------------------------------------- *)
(** sanity check: un terme hashconsé ne doit pas contenir le tag [Tag.default],
    et doit être tel que le re-hashconser donne le même terme (physiquement). *)
let is_hashconsed (t : term) : bool =
  let rec checktags : term -> bool = function
    | Tfun (tag, _, terms) -> tag <> Tag.default && List.for_all checktags terms
  in
  if checktags t then
    ( assert (hashcons t == t); (* egalité physique *)
      true )
  else false

(* ------------------------------------------------------------------- *)
module S = struct
  let kind  = `Standard
  let tfun  = tfun
  let equal = (=)               (* deep equality *)
end

module H = struct
  let kind  = `Hashconsing
  let tfun  = HashCons.fun_hashcons
  let equal = (==)              (* physical equality *)
end

(* ------------------------------------------------------------------- *)
(* sans hashconsing *)
module Standard = Make(S)

(* ------------------------------------------------------------------- *)
(** Question 8 -- numerical results *)
    
(** numerical results here to have results of question in order
    (everything is mixed because of modules). *)
let _res8 =
  let terms = read_terms (path ^ "q8.txt") in
  let doit t =
    let t' = hashcons t in
    let s1, s2 = (size t', size_mem t') in
    assert (size t = s1 && s1 >= s2);
    (s1, s2)
  in
  let pp fmt (d1,d2) = Format.fprintf fmt "(%d,%d)" d1 d2 in

  Format.printf "8.a %a@." pp (doit (List.nth terms 0));
  Format.printf "8.b %a@." pp (doit (List.nth terms 1));
  Format.printf "8.c %a@." pp (doit (List.nth terms 2));
  Format.printf "8.d %a@." pp (doit (List.nth terms 3));
  Format.printf "@."

(* ------------------------------------------------------------------- *)
(* avec hashconsing *)
module Hashcons = Make(H)

(* ------------------------------------------------------------------- *)
(** Question 9 *)

let _res9 =
  let terms = read_terms (path ^ "tq9.txt") in
  let rules = read_rules (path ^ "rq9.txt") in
  let doit t =
    let t', steps = Hashcons.rewriteN (hashcons t) ~rules in
    (size_mem t', steps)
  in
  let pp fmt (d1,d2) = Format.fprintf fmt "(%d,%d)" d1 d2 in

  Format.printf "9.a %a@." pp (doit (List.nth terms 0));
  Format.printf "9.b %a@." pp (doit (List.nth terms 1));
  Format.printf "9.c %a@." pp (doit (List.nth terms 2));
  Format.printf "9.d %a@." pp (doit (List.nth terms 3));
  Format.printf "@."

(* ------------------------------------------------------------------- *)
(** Question 10 *)

let rewrite1_all (t : term) (rules : rule list) : term list =     
  let rec doit t : term list =
    let Tfun (_, f, args) = t in
    let terms_head =
      (* itére sur les règles *)
      List.filter_map (fun (rl, rr) ->
          match Hashcons.find_match t rl with
          | None   -> None
            
          (* [s] tel que [t = subst s rl], on stocke [subst s rr] *)
          | Some s -> Some (Hashcons.subst s rr)
        ) rules
    in
    (* on récurse pour trouver des matches dans les sous-termes de [t] *)
    let terms_sub =
      let list_args' = doit_sub ~left:[] ~right:args ~acc:[] in
      List.map (fun args' -> HashCons.fun_hashcons f args') list_args'
    in
    terms_head @ terms_sub

  (* [left] in reverse order *)
  and doit_sub
      ~(left : term list) ~(right : term list) ~(acc : term list list)
    : term list list
    =
    match right with
    | [] -> acc
    | arg :: right ->
      let list_arg' = doit arg in
      let acc =
        List.fold_left (fun acc arg' ->
            let args' = List.rev_append left (arg' :: right) in
            args' :: acc
          ) acc list_arg'
      in
      doit_sub ~left:(arg :: left) ~right ~acc
  in
  (* Table de hachage utilisé comme un set pour enlever les doublons.
     On aurait aussi pu enlever les doublons lors des appels récursif. *)
  let hashtbl = Hashtbl.create 65536 in
  List.iter (fun x -> Hashtbl.add hashtbl (get_tag x) x) (doit t);
  Hashtbl.fold (fun _ t acc -> t :: acc) hashtbl []

(* ------------------------------------------------------------------- *) 
let longest (t : term) (rules : rule list) : int * term list =
  (* la longueur du plus grand chemin au départ d'un terme *)
  let hashtbl : (_, int * term list) Hashtbl.t = Hashtbl.create 65536 in
  
  let rec longest (t : term) : int * term list =
    try Hashtbl.find hashtbl (get_tag t) with
    | Not_found ->
      let nexts = rewrite1_all t rules in
      (* on a [t → t'] *)
      let max, path =
        (* séquence la plus longue au départ d'un successeur *)
        List.fold_left (fun (max,path) t' ->
            let len', path' = longest t' in
            if max < len' then (len', path') else (max, path)
          ) (0, []) nexts
      in
      let max, path = max + 1, t :: path in
      Hashtbl.add hashtbl (get_tag t) (max, path);
      (max, path)
  in
  let max, path = longest t in

  (* la nombre de ré-écriture est la longueur du chemin - 1 *)
  max - 1, path

let _res10 =
  let terms = read_terms (path ^ "tq10.txt") in
  let rules = read_rules (path ^ "rq10.txt") in
  let doit t =
    let len, l = longest (hashcons t) rules in
    assert (List.length l - 1 = len);
    len
  in
  let pp fmt d = Format.fprintf fmt "%d" d in

  Format.printf "10.a %a@." pp (doit (List.nth terms 0));
  Format.printf "10.b %a@." pp (doit (List.nth terms 1));
  Format.printf "10.c %a@." pp (doit (List.nth terms 2));
  Format.printf "10.d %a@." pp (doit (List.nth terms 3));
  Format.printf "@."

